/*----------------------------------------------------------------------
  File    : cluster2.c
  Contents: cluster and cluster set management (update functions)
  Author  : Christian Borgelt
  History : 2001.09.05 file created as cluster1.c
            2001.09.12 function cls_init() completed
            2001.09.15 first version of fuzzy c-means completed
            2001.09.16 hard c-means clustering algorithm added
            2002.09.09 neural network update methods added
            2003.01.29 some cleanup of neural network methods
            2003.01.31 initialization for resilient method changed
            2003.02.18 multiple init. in mode CLS_POINTS removed
            2003.06.07 cluster size adaptation added
            2003.10.29 bug in function cls_init() fixed (CLS_POINTS)
            2004.02.22 size computation for spherical clusters modified
            2004.02.23 bug in function quick() fixed
            2004.02.29 some bugs in cluster size computation fixed
            2004.03.01 update functions moved to this file
            2004.03.02 shape and size regularization added
            2004.03.19 weight/prior regularization added
            2004.04.12 competitive learning function completed
            2004.04.13 regularization adapted for competitive learning
            2004.04.14 treatment of (almost) empty clusters improved
            2004.04.15 update function improved (loop moved)
            2004.04.23 some numeric problems with high-dim. data solved
            2004.04.26 more numeric problems with high-dim. data solved
            2004.04.27 upper learning rate bound added (alt. est. step)
            2004.04.30 rescaling removed from parameter est. functions
            2004.07.13 normalization of center vectors added
            2004.07.15 bug in center normalization for complrn() fixed
            2004.07.28 update of centers only added to cls_update()
            2004.08.14 bug in initialization with CLS_POINTS fixed
            2004.08.18 first version of backpropagation functions added
            2007.11.22 eigenvalue sorting added for covariance mode
            2007.11.27 unit centers (on unit hyper-sphere) corrected
            2008.02.29 redesign for feature weighting completed
            2008.04.08 regularization functions adapted to new version
            2008.04.11 neural network inspired update debugged
            2008.04.14 neural network inspired update improved
            2008.04.21 modify only covariance changes towards zero
            2008.04.22 momentum term update bounded by maximal change
            2008.04.25 covariance matrix checks improved in neural()
            2008.07.14 bug in function cls_update() fixed (regshape())
            2010.04.08 special handling of squares removed
            2013.09.02 bug in cls_init() fixed (setting of c->size)
            2021.08.25 bug in function neural() fixed (update function)
            2021.09.01 bug in functions cls_bkprop() and rbfgrad() fixed
----------------------------------------------------------------------*/
#include <stdlib.h>
#include <float.h>
#include <math.h>
#include <assert.h>
#include "cluster.h"

/*----------------------------------------------------------------------
  Preprocessor Definitions
----------------------------------------------------------------------*/
#define MINVAR      1e-48       /* minimal variance */
#define MINDET      1e-48       /* minimal determinant */
#define MAXDET      1e+48       /* maximal determinant */
#define MINWEIGHT   1e-12       /* minimal cluster weight */

/*----------------------------------------------------------------------
  Type Definitions
----------------------------------------------------------------------*/
typedef MATRIX* MATADDFN (MATRIX *mat, const double *vec, double wgt);
typedef double  UPDATEFN (CLSET *clset, double  grd,
                          double prv,   double *chg);

/*----------------------------------------------------------------------
  Auxiliary Functions from cluster1.c
----------------------------------------------------------------------*/
extern void cls_combine (CLSET *clset);
extern void cls_distrib (CLSET *clset);
extern void cls_vars    (CLSET *clset, int eigen);
extern void cls_gauge   (CLSET *clset);
extern void cls_resize  (CLSET *clset, int decomp);
extern void cls_ftwgts  (CLSET *clset);
extern void cls_invert  (CLSET *clset);
extern void cls_reinit  (CLSET *clset);

/*----------------------------------------------------------------------
  Gradient Function
----------------------------------------------------------------------*/

static void gradient (CLSET *clset)
{                               /* --- gradient based update */
  /* ... to be done ... */
}  /* gradient() */

/*----------------------------------------------------------------------
  Alternating Optimization Function
----------------------------------------------------------------------*/

static void altopt (CLSET *clset)
{                               /* --- alternating optimization */
  DIMID   i;                    /* loop variable */
  CLUSTER *c;                   /* to traverse the clusters */
  MATRIX  *mat;                 /* exchange buffer for matrices */
  double  w, sum = 0;           /* (sum of) cluster weights */

  assert(clset);                /* check the function argument */
  /* --- check and sum the weights --- */
  c = clset->cls;               /* get the cluster array */
  for (i = 0; i < clset->clscnt; i++) {
    w = mat_getwgt(c[i].smp);   /* traverse the clusters */
    if (w >= MINWEIGHT) {                       c[i].ok = -1; }
    else { mat_setwgt(c[i].smp, w = MINWEIGHT); c[i].ok =  0; }
    sum += w;                   /* check and sum the cluster weights */
    mat = c[i].smp; c[i].smp = c[i].cov; c[i].cov = mat;
    c[i].ctr = mat_vector(c[i].cov); /* exchange covariance matrix */
    c[i].sum = mat_vector(c[i].smp); /* and aggregation matrix and */
  }                                  /* get corresponding buffers */
  sum = (sum > 0) ? 1/sum : 1;  /* compute the normalization factor */

  /* --- compute new centers and (co)variances --- */
  if      (clset->type & CLS_COVARS) { /* if adaptable covariances */
    for (i = 0; i < clset->clscnt; i++) {
      if (c[i].ok) mat_covarx(c[i].cov, c[i].cov, 1);
      else mat_copyx(c[i].cov, c[i].smp, MAT_VECTOR|MAT_UPPER);
      mat_mulwgt(c[i].cov,sum); /* compute new covariance matrix */
    } }                         /* and normalize the weights */
  else if (clset->type & CLS_VARS) {   /* if adaptable variances */
    for (i = 0; i < clset->clscnt; i++) {
      if (c[i].ok) mat_varx(c[i].cov, c[i].cov, 1);
      else mat_copyx(c[i].cov, c[i].smp, MAT_VECTOR|MAT_DIAG);
      mat_mulwgt(c[i].cov,sum); /* compute new variances */
    } }                         /* and normalize the weights */
  else if (clset->type & CLS_SIZE) {   /* if adaptable size */
    for (i = 0; i < clset->clscnt; i++) {
      if (c[i].ok) mat_isovarx(c[i].cov, c[i].cov, 1);
      else mat_copyx(c[i].cov, c[i].smp, MAT_VECTOR|MAT_CORNER);
      mat_mulwgt(c[i].cov,sum); /* compute new cluster size */
    } }                         /* and normalize the weight */
  else {                        /* if no shape or size flags set */
    for (i = 0; i < clset->clscnt; i++) {
      if (c[i].ok) mat_mean(c[i].ctr, c[i].cov);
      else mat_copyx(c[i].cov, c[i].smp, MAT_VECTOR);
      mat_mulwgt(c[i].cov,sum); /* only compute new cluster centers */
    }                           /* and normalize the weights */
  }                             /* (copy old values if necessary) */

  /* --- combine (co)variances --- */
  if ((clset->type &  CLS_JOINT)
  &&  (clset->type & (CLS_COVARS|CLS_VARS|CLS_SIZE)))
    cls_combine(clset);         /* combine the (co)variances */
}  /* altopt() */

/*----------------------------------------------------------------------
  Competitive Learning Function
----------------------------------------------------------------------*/

static void complrn (CLSET *clset)
{                               /* --- competitive learning */
  int     m;                    /* matrix mode */
  DIMID   i, n;                 /* loop variable */
  CLUSTER *c;                   /* to traverse the clusters */
  double  w, sum;               /* (sum of) cluster weights */
  double  lrn, dec, eta;        /* learning rate and related values */
  MATRIX  *mat;                 /* exchange buffer */

  assert(clset);                /* check the function argument */

  /* --- compute new centers --- */
  c   = clset->cls;             /* get the cluster array */
  dec = clset->decay[0];        /* get the decay parameter and */
  lrn = clset->rates[0];        /* compute the next learning rate */
  if ((dec > 0) && (dec < 1)) lrn *= pow(dec, clset->steps);
  else if          (dec < 0)  lrn *= pow(clset->steps+1, dec);
  for (i = 0; i < clset->clscnt; i++) {
    w = mat_getwgt(c[i].smp);   /* traverse the clusters */
    w = (w > 1) ? lrn/w : lrn;  /* compute the learning rate */
    mat = c[i].smp; c[i].smp = c[i].cov; c[i].cov = mat;
    c[i].ctr = mat_vector(c[i].cov); /* exchange covariance matrix */
    c[i].sum = mat_vector(c[i].smp); /* and aggregation matrix */
    vec_add(c[i].ctr, clset->inpcnt, c[i].sum, w, c[i].ctr);
  }                             /* compute new center coordinates */
  /* Note that c[i].sum is the aggregate of the difference vectors to */
  /* the cluster centers and *not* the aggregate of the data vectors  */
  /* as for alternating estimation/fuzzy clustering. Hence there is   */
  /* no decay factor for the old cluster center c[i].ctr (cf. below). */

  /* --- compute new (co)variances --- */
  if (clset->type & (CLS_COVARS|CLS_VARS|CLS_SIZE)) {
    dec = clset->decay[1];      /* get the decay parameter and */
    lrn = clset->rates[1];      /* compute the next learning rate */
    if ((dec > 0) && (dec < 1)) lrn *= pow(dec, clset->steps);
    else if          (dec < 0)  lrn *= pow(clset->steps+1, dec);
    m = (clset->type & CLS_COVARS) ? MAT_UPPER : MAT_DIAG;
    if (!(clset->type & CLS_JOINT))
      n = clset->clscnt;        /* default: process all clusters */
    else {                      /* if to combine (co)variances */
      for (i = 0; i < clset->clscnt; i++)
        mat_addx(c[0].cov, c[0].cov, 1, c[i].cov, m);
      n = 1;                    /* combine them in first cluster and */
    }                           /* then process only this cluster */
    for (i = 0; i < n; i++) {   /* traverse the (relevant) clusters */
      w = mat_getwgt(c[i].cov); /* compute learning rate and compl. */
      if (w >= 1) { eta = lrn/w; dec = 1 -lrn;   }
      else        { eta = lrn;   dec = 1 -lrn*w; }
      if (clset->type & (CLS_COVARS|CLS_VARS)) {
        mat_mulsx(c[i].cov, c[i].cov, eta, m);
        if (dec > 0) mat_addx(c[i].cov, c[i].cov, dec, c[i].smp, m); }
      else {                    /* if adaptable (co)variances */
        w = eta *(mat_trace(c[i].cov) /(double)clset->inpcnt);
        if (dec > 0) w += dec *mat_get(c[i].smp, 0, 0);
        mat_set(c[i].cov, 0, 0, w);
      }                         /* if only isoptropic variance, */
    }                           /* use specialized procedure */
  }                             /* (update (co)variances/sizes) */

  /* --- compute new weights --- */
  if (clset->type & CLS_WEIGHT){/* if to use cluster weights */
    dec = clset->decay[2];      /* get the decay parameter and */
    lrn = clset->rates[2];      /* compute the next learning rate */
    if ((dec > 0) && (dec < 1)) lrn *= pow(dec, clset->steps);
    else if          (dec < 0)  lrn *= pow(clset->steps+1, dec);
    dec = 1 -lrn; sum = 0;      /* traverse the clusters */
    for (i = 0; i < clset->clscnt; i++) {
      w = lrn *mat_getwgt(c[i].cov) +dec *mat_getwgt(c[i].smp);
      if (!(w >= MINWEIGHT)) w = MINWEIGHT;
      sum += mat_setwgt(c[i].cov, w);    /* adapt the weights and */
    }                           /* sum them for the normalization */
    sum = (sum > 0) ? 1/sum : 0;/* compute the normalization factor */
    for (i = 0; i < clset->clscnt; i++)
      mat_mulwgt(c[i].cov, sum);
  }                             /* normalize the new cluster weights */

  clset->steps++;               /* count the update step */
}  /* complrn() */

/*----------------------------------------------------------------------
  Gradient Descent for Radial Basis Function Networks
----------------------------------------------------------------------*/

static void rbfgrad (CLSET *clset)
{                               /* --- gradient descent for RBFs */
  DIMID   i;                    /* loop variable */
  CLUSTER *c;                   /* to traverse the clusters */
  double  lrc, lrv, t, v;       /* learning rates, buffers */

  assert(clset);                /* check the function argument */
  c   =  clset->cls;            /* get the cluster array */
  lrc =  clset->rates[0];       /* get the center   learning rate */
  lrv = -clset->rates[1];       /* and the (co)var. learning rate */
  for (i = 0; i < clset->clscnt; i++) {  /* traverse the clusters */
    vec_add(c[i].ctr, clset->inpcnt, c[i].ctr, lrc, c[i].sum);
    if (clset->type & CLS_COVARS) {  /* if adaptable covariances */
      mat_mulsx(c[i].smp, c[i].smp, lrv, MAT_UPPER);
      mat_addx(c[i].smp, c[i].cov, 1, c[i].smp, MAT_UPPER);
      if (mat_chinvx(c[i].inv, c[i].smp) >= 0)
        mat_copyx(c[i].cov, c[i].smp, MAT_UPPER); }
    else if (clset->type & CLS_VARS) { /* if adaptable variances */
      mat_diamuls(c[i].smp, lrv);
      mat_diaadd(c[i].smp, 1, c[i].cov);
      if (mat_diainv(c[i].inv, c[i].smp) >= 0)
        mat_copyx(c[i].cov, c[i].smp, MAT_DIAG); }
    else if (clset->type & CLS_SIZE) {
      v = mat_get(c[i].cov, 0,0);    /* if adaptable size */
      t = v +lrv *mat_get(c[i].smp, 0, 0);
      mat_set(c[i].cov, 0, 0, (t > 0) ? t : v);
    }                           /* compute new (co)variances */
  }                             /* or new isotropic variance */
  clset->steps++;               /* count the update step */
}  /* rbfgrad() */

/*----------------------------------------------------------------------
  Neural Network Update Functions
----------------------------------------------------------------------*/

static double standard (CLSET *clset,
                        double grd, double prv, double *chg)
{                               /* --- standard update */
  return grd;                   /* (this function is actually unused) */
}  /* standard() */

/*--------------------------------------------------------------------*/

static double expand (CLSET *clset,
                      double grd, double prv, double *chg)
{                               /* --- update expanded by a factor */
  return clset->growth *grd;
}  /* expand() */

/*--------------------------------------------------------------------*/

static double momentum (CLSET *clset,
                        double grd, double prv, double *chg)
{                               /* --- update with momentum term */
  double g, d;                  /* temporary buffers */

  *chg = grd +*chg *clset->moment; /* compute the parameter change */
  if (clset->rbfgrad)           /* if gradient descent for RBFs, */
    return *chg;                /* return the computed change */
  g = fabs(grd); d = fabs(*chg);
  if (d < g) return *chg = grd; /* update at least with the gradient */
  g *= clset->maxchg;           /* constrain the parameter change */
  if (d > g) return *chg = grd *clset->maxchg;
  return *chg;                  /* return the parameter change */
}  /* momentum() */

/*--------------------------------------------------------------------*/

static double adaptive (CLSET *clset,
                        double grd, double prv, double *chg)
{                               /* --- self-adaptive learning rate */
  prv *= grd;                   /* compute relative change directions */
  if      (prv < 0) {           /* if gradients have opposite signs */
    *chg *= clset->shrink;      /* decrease the learning rate */
    if (*chg < clset->minchg) *chg = clset->minchg; }
  else if (prv > 0) {           /* if gradients have the same sign */
    *chg *= clset->growth;      /* increase the learning rate */
    if (*chg > clset->maxchg) *chg = clset->maxchg;
  }                             /* clamp to learning rate range */
  return *chg *grd;             /* return the parameter change */
}  /* adaptive() */

/*--------------------------------------------------------------------*/

static double resilient (CLSET *clset,
                         double grd, double prv, double *chg)
{                               /* --- resilient backpropagation */
  double g;                     /* temporary buffer */

  if (*chg == 0) {              /* if no step has been carried out, */
    *chg = fabs(grd); return grd; }        /* initialize the change */
  prv *= grd;                   /* compute relative change directions */
  if      (prv > 0)             /* if gradients have the same sign, */
    *chg *= clset->growth;      /* increase the change value */
  else if (prv < 0)             /* if gradients have opposite signs, */
    *chg *= clset->shrink;      /* decrease the change value */
  if (clset->rbfgrad) {         /* if gradient descent for RBFs */
    if (*chg < clset->minchg) *chg = clset->minchg;
    if (*chg > clset->maxchg) *chg = clset->maxchg; }
  else {                        /* if clustering update */
    g = fabs(grd);              /* update at least with the gradient */
    if (*chg < g) *chg = g;     /* (do at least standard update) */
    g *= clset->maxchg;         /* restrict the parameter change */
    if (*chg > g) *chg = g;     /* to maxchg times the gradient */
  }
  return (grd < 0) ? -*chg : *chg; /* return the parameter change */
}  /* resilient() */

/*--------------------------------------------------------------------*/

static double quick (CLSET *clset,
                     double grd, double prv, double *chg)
{                               /* --- quickprop analog */
  double g, d;                  /* temporary buffers */

  if (prv == 0)                 /* if this is the first update, */
    return *chg = grd;          /* do a standard update step */
  d = prv -grd;                 /* compute the gradient change */
  if (*chg *d <= 0)             /* if the parabola opens downwards, */
    return *chg = grd;          /* do a standard update step */
  g = prv *clset->maxfac;       /* compute limit for new gradient */
  if (prv > 0) {                /* if previous gradient was positive */
    if (grd < g)                /* compute the factor for a jump */
      *chg *= grd /d;           /* to the minimum (apex of parabola) */
    else                        /* if the growth factor would become */
      *chg *= clset->growth; }  /* too large, use the maximal factor */
  else {                        /* if previous gradient was negative */
    if (grd > g)                /* compute the factor for a jump */
      *chg *= grd /d;           /* to the minimum (apex of parabola) */
    else                        /* if the growth factor would become */
      *chg *= clset->growth;    /* too large, use the maximal factor */
  }                             /* (avoid jumps that are too large) */
  d = fabs(*chg);               /* get the absolute change */
  if (clset->rbfgrad) {         /* if gradient descent for RBFs */
    if (d < clset->minchg)      /* update at least with min. change */
      *chg = (grd < 0) ? -clset->minchg : clset->minchg;
    if (d > clset->maxchg)      /* update at most  with max. change */
      *chg = (grd < 0) ? -clset->maxchg : clset->maxchg; }
  else {                        /* if clustering update */
    g = fabs(grd);              /* update at least with the gradient */
    if (d < g) return *chg = grd;
    g *= clset->maxchg;         /* at most with maxchg x grapdient */
    if (d > g) return *chg = grd *clset->maxchg;
  }                             /* (constrain the parameter change) */
  return *chg;                  /* return the parameter change */
}  /* quick() */

/*--------------------------------------------------------------------*/

static UPDATEFN *updatefn[] = {
  /* CLS_NONE       0x00 */  standard,
  /* CLS_EXPAND     0x10 */  expand,
  /* CLS_MOMENTUM   0x20 */  momentum,
  /* CLS_ADPATIVE   0x30 */  adaptive,
  /* CLS_RESILIENT  0x40 */  resilient,
  /* CLS_QUICK      0x50 */  quick,
};                              /* list of update functions */

/*--------------------------------------------------------------------*/

static void neural (CLSET *clset, int eigen)
{                               /* --- neural network inspired update */
  DIMID    i, k, n, m;          /* loop variables, buffers */
  int      adapt;               /* flag for adaptive learning rate */
  CLUSTER  *c;                  /* to traverse the clusters */
  double   *s, *z, *x, *b;      /* to access the vectors */
  double   v, w, d, g, p;       /* temporary buffers */
  UPDATEFN *update;             /* parameter update function */

  assert(clset                  /* check the function argument */
  &&    (clset->method & CLS_MODIFIER));
  c = clset->cls;               /* get the cluster array */
  m = clset->method & CLS_MODIFIER;
  adapt  = (m == CLS_ADAPTIVE); /* get the adaptive update flag */
  update = updatefn[m >> 16];   /* and the parameter update function */

  /* --- update the cluster centers --- */
  for (n = 0; n < clset->clscnt; n++) {
    s = c[n].ctr; z = c[n].sum; /* traverse the clusters */
    x = c[n].chc; b = c[n].grc; /* get the update values */
    for (k = clset->inpcnt; --k >= 0; ) {
      d    = s[k] -z[k];        /* traverse the dimensions */
      s[k] = z[k] +update(clset, d, b[k], x+k);
      b[k] = d;                 /* compute the current gradient, */
    }                           /* update the center coordinate, */
  }                             /* and store the current gradient */
  if (!(clset->type   & (CLS_COVARS|CLS_VARS|CLS_SIZE))
  ||  !(clset->method &  CLS_MODVAR))
    return;                     /* check for (co)variance adaptation */

  /* --- update only variances/isotropic variance --- */
  if (!(clset->type & CLS_COVARS)) {
    m = (clset->type & CLS_VARS)  ? clset->inpcnt : 1;
    n = (clset->type & CLS_JOINT) ? 1 : clset->clscnt;
    while (--n >= 0) {          /* traverse the relevant clusters */
      for (i = 0; i < m; i++) { /* traverse the variances */
        v = mat_get(c[n].smp, i, i);  /* get the old and the */
        w = mat_get(c[n].cov, i, i);  /* new value of a variance, */
        g = mat_get(c[n].grv, i, i);  /* the previous gradient */
        p = mat_get(c[n].chv, i, i);  /* and the previous change */
        v += update(clset, d = w-v, g, &p);
        if (v < MINVAR) {       /* compute the new variance */
          v = w; p = (adapt) ? 1 : d; }
        mat_set(c[n].cov, i, i, v);   /* store the new variance, */
        mat_set(c[n].grv, i, i, d);   /* its (unmodified) gradient */
        mat_set(c[n].chv, i, i, p);   /* and the updated parameter */
      }                         /* (the parameter has been reset */
    } }                         /* if the variance got negative) */

  /* --- update full covariance matrix --- */
  else {                        /* if there are covariances */
    n = (clset->type & CLS_JOINT) ? 1 : clset->clscnt;
    while (--n >= 0) {          /* traverse the relevant clusters */
      for (i = clset->inpcnt; --i >= 0; ) {
        for (k = clset->inpcnt; --k >= i; ) {
          v = mat_get(c[n].smp, i, k);    /* get a (co)variance */
          d = mat_get(c[n].cov, i, k) -v; /* and its change */
          g = mat_get(c[n].grv, i, k);    /* get previous gradient */
          p = mat_get(c[n].chv, i, k);    /* and previous change */
          v += update(clset, d, g, &p); /* update the (co)variance */
          mat_set(c[n].mat, i, k, v);     /* store the value of the */
          mat_set(c[n].grv, i, k, d);     /* covari. matrix element */
          mat_set(c[n].chv, i, k, p);     /* and updated parameter */
        }                       /* (compute a new covariance matrix */
      }                         /* in the matrix buffer) */
      if (!eigen)               /* do Cholesky decomposition */
        k = mat_chdecom(c[n].inv, c[n].mat);
      else {                    /* do eigenvalue decomposition */
        mat_3dred(c[n].buf, c[n].buf +n, c[n].inv, c[n].mat, MAT_UPPER);
        k = mat_3dqli(c[n].dif, c[n].inv, c[n].buf, c[n].buf+n, n, 256);
        if (k == 0) {           /* if successful decomposition */
          mat_evsort(c[n].dif, c[n].inv, 1);
          k = (c[n].dif[0] < MINVAR) ? -1 : 0;
        }                       /* check for positive eigenvalues */
      }                         /* (positive definite matrix) */
      if (k == 0) {             /* if successful decomposition */
        mat_copyx(c[n].cov, c[n].mat, MAT_UPPER);
        c[n].dec = -1; }        /* copy the update result */
      else {                    /* if decomposition failed */
        if (!adapt) mat_addx(c[n].chv, c[n].cov,-1,c[n].smp, MAT_UPPER);
        else { p = 1; mat_init(c[n].chv, MAT_UPPER|MAT_VALUE, &p); }
      }                         /* on failure reinitialize */
    }                           /* the additional parameter */
  }                             /* (reset to consistent state) */
  /* If the update of the covariance matrix fails, the originally */
  /* computed covariance matrix is kept (standard update step).   */
}  /* neural() */

/*----------------------------------------------------------------------
  Regularization Functions
----------------------------------------------------------------------*/

static void regshape (CLSET *clset)
{                               /* --- regularize cluster shapes */
  DIMID   i, k, m;              /* loop variables, buffer */
  CLUSTER *c;                   /* to traverse the clusters */
  double  t, h, s, a;           /* temporary buffers */
  double  min, max;             /* minimum and maximum eigenvalue */

  assert(clset                  /* check the function argument */
  &&    (clset->inpcnt > 1)     /* and the regularization need */
  &&    (clset->type & (CLS_COVARS|CLS_VARS))
  &&    (clset->eigen || !(clset->type & CLS_COVARS))
  &&    ((clset->fwexp <= 0) || (clset->fwexp >= 1)));
  h = clset->regps[3]; s = h*h; /* get the regularization parameter */
  assert((h < -1) || (h > 0));  /* and check for the active range */
  m = clset->inpcnt;            /* get the number of dimensions */
  c = clset->cls;               /* and the cluster array */
  for (i = (clset->type & CLS_JOINT) ? 1 : clset->clscnt; --i >= 0; ) {
    if (h >= 0)                 /* if to apply standard version */
      a = s *c[i].var;          /* compute offset h^2 * \sigma^2 */
    else {                      /* if to apply alternative version */
      min = INFINITY; max = 0;  /* initialize the eigenvalue range */
      for (k = 0; k < m; k++) { /* traverse the eigenvalues */
        t = (clset->type & CLS_COVARS)
          ? c[i].dif[k] : mat_get(c[i].cov, k, k);
        if (t < min) min = t;   /* find the minimum and maximum */
        if (t > max) max = t;   /* of the eigenvalues (axes lengths) */
      }                         /* for the ratio computation */
      t = max -s *min;          /* compute numerator of fraction */
      if (!(t > 0)) continue;   /* and check against maximum ratio */
      a = t /(s -1);            /* compute the regularization value */
    }                           /* to ensure the maximum ratio */
    assert(a >= 0);             /* check the (computed) shift value */
    mat_diaadds(c[i].cov, a);   /* and add it to the diagonal */
    if (clset->type & CLS_COVARS) { /* if full covariance matrix */
      for (t = 1, k = m; --k >= 0; )
        t *= c[i].dif[k] += a;  /* update the eigenvalues */
      c[i].var = ((t >= MINDET) && (t <= MAXDET))
               ? pow(t,                   1.0/(double)m)
               : exp(vec_sumlog(c[i].dif, m) /(double)m); }
    else {                      /* if only variances */
      t = mat_diaprod(c[i].cov);/* compute the determinant */
      c[i].var = ((t >= MINDET) && (t <= MAXDET))
               ? pow(t,                   1.0/(double)m)
               : exp(mat_dialog(c[i].cov)    /(double)m);
    }                           /* compute the new isotropic variance */
  }
}  /* regshape() */

/*--------------------------------------------------------------------*/

static void regsize (CLSET *clset)
{                               /* --- regularize cluster sizes */
  DIMID   i;                    /* loop variable */
  CLUSTER *c;                   /* to traverse the clusters */
  double  min, max;             /* minimum and maximum size */
  double  s, t, a, b;           /* temporary buffers */

  assert(clset                  /* check the function argument */
  &&    (clset->type & CLS_SIZE) && !(clset->type & CLS_JOINT)
  &&    ((clset->fwexp <= 0) || (clset->fwexp == 1)));
  s = clset->regps[2];          /* get the scaling factor, */
  a = clset->regps[1] *0.5;     /* the variance exponent, and */
  b = clset->regps[0];          /* the regularization offset */
  if ((a <= 0) || (s == 0)      /* check whether regularization */
  || ((s == 1) && (b == 0)))    /* is actually to be done */
    return;                     /* (active parameter combination) */
  c = clset->cls;               /* get the cluster array */
  if (a == 1) {                 /* if to use the isotropic variance */
    for (i = 0; i < clset->clscnt; i++)
      c[i].d2 = c[i].var; }     /* simply copy the isotropic variance */
  else {                        /* if to use some power instead */
    for (i = 0; i < clset->clscnt; i++)
      c[i].d2 = pow(c[i].var, a);  /* compute the requested power */
  }                                /* of the isotropic variance */
  if (b < -1) {                 /* if alternative version */
    min = INFINITY; max = 0;    /* initialize the size range */
    for (i = 0; i < clset->clscnt; i++) {
      t = c[i].d2;              /* traverse the cluster sizes */
      if (t < min) min = t;     /* and determine their minimum */
      if (t > max) max = t;     /* and their maximum for the */
    }                           /* size ratio computation */
    t = max +b *min;            /* compute numerator of fraction, */
    if (!(t > 0)) return;       /* check against the maximum ratio, */
    b = t /(-1 -b);             /* and compute the necessary offset */
  }                             /* to ensure the maximum ratio */
  assert(b >= 0);               /* check the (computed) offset */
  t = 0;                        /* initialize the cluster size sum */
  if (s <= 0) {                 /* if to compute simplified version */
    if (b < 0) return;          /* check for a non-negative offset */
    for (i = 0; i < clset->clscnt; i++)
      c[i].d2 += b;             /* add given offset to the sizes */
    s = -s; }                   /* get the (re)scaling factor */
  else if (b < 0) {             /* if to equalize the sizes fully */
    for (i = 0; i < clset->clscnt; i++) {
      t += c[i].d2; c[i].d2 = 1; }   /* set unit cluster sizes and */
    s *= t /(double)clset->clscnt; } /* compute (re)scaling factor */
  else {                        /* if to equalize the sizes partially */
    for (i = 0; i < clset->clscnt; i++) { /* sum the cluster sizes */
      t += c[i].d2; c[i].d2 += b; }       /* and add given offset */
    s *= t/(t +(double)clset->clscnt *b);
  }                             /* compute the (re)scaling factor */
  a = 1/a;                      /* get the inverse variance exponent */
  for (i = 0; i < clset->clscnt; i++) {
    t = pow(s *c[i].d2, a);     /* compute the new cluster size */
    b = t /c[i].var; c[i].var = t;  /* and the rescaling factor */
    if (!(clset->type & CLS_COVARS))
      mat_diamuls(c[i].cov, b); /* rescale the variances */
    else {                      /* if adaptable covariances */
      mat_mulsx(c[i].cov, c[i].cov, b, MAT_UPPER);
      if (clset->eigen) vec_muls(c[i].dif, clset->inpcnt, c[i].dif, b);
      else mat_mulsx(c[i].inv, c[i].inv, sqrt(b), MAT_LOWER);
    }                           /* rescale the eigenvalues or */
  }                             /* the Cholesky decomposition */
}  /* regsize() */

/*--------------------------------------------------------------------*/

static void regwgts (CLSET *clset)
{                               /* --- regularize cluster weights */
  DIMID   i;                    /* loop variable */
  CLUSTER *c;                   /* to traverse the clusters */
  double  min, max;             /* minimum and maximum weight */
  double  a, t, w;              /* regularization parameter, buffers */

  assert(clset                  /* check the function arguments */
  &&    (clset->type & CLS_WEIGHT) && !(clset->type & CLS_JOINT));
  a = clset->regps[4];          /* get the regularization parameter */
  c = clset->cls;               /* and the clusters array */
  assert((a < -1) || (a > 0));  /* and check for the active range */
  if (a < 0) {                  /* if alternative reg. version, */
    min = INFINITY; max = 0;    /* initialize the weight range */
    for (i = 0; i < clset->clscnt; i++) {
      w = mat_getwgt(c[i].cov); /* traverse the clusters and */
      if (w < min) min = w;     /* determine the weight range */
      if (w > max) max = w;     /* (minimum and maximum weight) */
    }
    t = max +a *min;            /* compute numerator of fraction, */
    if (!(t > 0)) return;       /* check against the maximum ratio, */
    a = t /(-1 -a);             /* and compute the necessary offset */
  }
  assert(a >= 0);               /* check the (computed) offset */
  t = 1 /(1 +a *(double)clset->clscnt);/* compute norm. factor */
  for (i = 0; i < clset->clscnt; i++) {
    w = mat_getwgt(c[i].cov);   /* traverse the clusters */
    mat_setwgt(c[i].cov, t *(w + a));
  }                             /* compute new cluster weights */
}  /* regwgts() */

/*--------------------------------------------------------------------*/

static void regctrs (CLSET *clset)
{                               /* --- regularize the centers */
  DIMID   i;                    /* loop variable */
  CLUSTER *c;                   /* to traverse the clusters */

  assert(clset                  /* check the function argument */
  &&    (clset->method & CLS_UNIT));
  c = clset->cls;               /* get the clusters array */
  if (clset->method & CLS_UNIT){/* make center of unit length */
    for (i = 0; i < clset->clscnt; i++)
      vec_unitlen(c[i].ctr, c[i].ctr, clset->inpcnt);
  }                             /* (enforce center constraints) */
}  /* regctrs() */

/*----------------------------------------------------------------------
  Main Functions
----------------------------------------------------------------------*/

void cls_init (CLSET *clset, int mode, RANDFN randfn, const double *vec)
{                               /* --- initialize a set of clusters */
  DIMID   i, k, n;              /* loop variables */
  CLUSTER *c;                   /* to traverse the clusters */
  NSTATS  *nst;                 /* simple numerical statistics */
  double  *z, *b;               /* cluster center, buffer */
  double  x, d, m;              /* buffers for computations */

  assert(clset && randfn);      /* check the function arguments */
  clset->setup = 0;             /* cluster set needs setup */
  c = clset->cls;               /* get the cluster array */

  /* --- initialize other parameters --- */
  if ((mode != CLS_POINTS)      /* if not called second time */
  ||  (clset->init <= 0) || (clset->init >= clset->clscnt)) {
    clset->type = CLS_CENTER;   /* only centers are initialized */
    for (k = 0; k < clset->clscnt; k++)
      mat_diasetx(c[k].cov, c[k].size);
  }                             /* set covariance matrix of clusters */

  /* --- initialize cluster centers --- */
  nst = clset->nst;             /* get the numerical statistics */
  switch (mode) {               /* evaluate the initialization mode */

    case CLS_CENTER:            /* -- center of the data space */
      z = c[0].ctr;             /* compute the center */
      nst_center(nst, z);       /* of the data space */
      for (k = 1; k < clset->clscnt; k++)
        vec_copy(c[k].ctr, z, clset->inpcnt);
      break;                    /* copy the center to all clusters */

    case CLS_UNIFORM:           /* -- uniformly distributed */
    default:                    /* (this is also the default) */
      nst_spans(nst, b = clset->vec);  /* get the value spans */
      for (k = 0; k < clset->clscnt; k++)
        for (i = 0; i < clset->inpcnt; i++)
          c[k].ctr[i] = nst_min(nst, i) +b[i] *randfn();
      break;                    /* set to random points in data space */

    case CLS_DIAG:              /* -- diagonal of the data space */
    case CLS_LATIN:             /* -- latin hypercube sampling */
      for (i = clset->inpcnt; --i >= 0; ) {
        m =  nst_min(nst, i);   /* compute value decrement */
        d = (nst_max(nst, i) -m) / (double)clset->clscnt;
        x = m +0.5*d;           /* compute first value */
        for (k = 0; k < clset->clscnt; k++) {
          c[k].ctr[i] = x; x += d; }
      }                         /* set equally spaced values */
      if (mode != CLS_LATIN)    /* if only to set the diagonal, */
        break;                  /* stop at this point */

    /* case CLS_LATIN: */       /* -- latin hypercube sampling */
      for (n = clset->clscnt; --n > 0; ) {
        for (i = clset->inpcnt; --i >= 0; ) {
          k = (DIMID)((double)(n+1) *randfn());
          if      (k > n) k = n;   /* compute a random index in */
          else if (k < 0) k = 0;   /* the remaining set of centers */
          x           = c[k].ctr[i];
          c[k].ctr[i] = c[n].ctr[i];
          c[n].ctr[i] = x;      /* exchange the i-th elements of the */
        }                       /* k-th and the n-th cluster center */
      } break;                  /* (shuffle dimensions independently) */

    case CLS_POINTS:            /* -- given points in the data space */
      if (vec) nst_norm(nst, vec, clset->vec);
      vec = clset->vec;         /* scale given vector to the buffer */
      if (clset->init >= clset->clscnt)
        clset->init = 0;        /* if all clusters are init., restart */
      vec_copy(c[clset->init].ctr, vec, clset->inpcnt);
      clset->init++;            /* copy the given vector and */
      break;                    /* count the initialized cluster */
  }

  /* --- scale the coordinates --- */
  if (mode != CLS_POINTS)       /* normalize center coordinates */
    for (k = 0; k < clset->clscnt; k++)
      nst_norm(nst, c[k].ctr, c[k].ctr);
}  /* cls_init() */

/*--------------------------------------------------------------------*/

void cls_jitter (CLSET *clset, double range, RANDFN randfn)
{                               /* --- add jitter to cluster centers */
  DIMID  i, k;                  /* loop variables */
  double *c;                    /* to access the cluster centers */

  assert(clset && randfn);      /* check the function arguments */
  if (range <= 0) return;       /* if no range is given, abort */
  for (k = 0; k < clset->clscnt; k++) {
    c = clset->cls[k].ctr;      /* traverse the cluster centers */
    for (i = 0; i < clset->inpcnt; i++)
      c[i] += (2 *randfn() -1) *range;
  }                             /* add a random offset to all values */
}  /* cls_jitter() */

/*--------------------------------------------------------------------*/

void cls_method (CLSET *clset, int method)
{                               /* --- set parameter update method */
  DIMID   i;                    /* loop variable */
  CLUSTER *c;                   /* to traverse the clusters */
  double  t;                    /* initialization value */

  assert(clset                  /* check the function arguments */
  &&    ((method & CLS_METHOD)   >= CLS_GRADIENT)
  &&    ((method & CLS_METHOD)   <= CLS_HYBRID)
  &&    ((method & CLS_MODIFIER) >= CLS_NONE)
  &&    ((method & CLS_MODIFIER) <= CLS_C2CALL));
  clset->method = method;       /* note the parameter update method */
  method &= CLS_MODIFIER;       /* get the update modifier */
  if (method > CLS_EXPAND) {    /* if one of the higher methods */
    t = (method == CLS_ADAPTIVE) ? 1 : 0;
    c = clset->cls;             /* get cluster array and adapt. flag */
    for (i = 0; i < clset->clscnt; i++) {
      mat_init(c[i].chv, MAT_FULL|MAT_VECTOR|MAT_VALUE, &t);
      mat_init(c[i].grv, MAT_FULL|MAT_VECTOR|MAT_ZERO,  NULL);
    }                           /* initialize the change matrix */
  }                             /* and the buffer matrix */
}  /* cls_method() */

/*--------------------------------------------------------------------*/

void cls_regular (CLSET *clset, const double *params)
{                               /* --- set regularization parameters */
  assert(clset);                /* check the function arguments */
  if (!params) return;          /* if regularization parameters */
  clset->regps[0] = params[0];  /* are given, copy them */
  clset->regps[1] = params[1];  /* to the cluster set */
  clset->regps[2] = params[2];
  clset->regps[3] = params[3];
  clset->regps[4] = params[4];
}  /* cls_regular() */

/*--------------------------------------------------------------------*/

void cls_lrate (CLSET *clset, const double *rates, const double *decays)
{                               /* --- set learning rate parameters */
  assert(clset);                /* check the function arguments */
  if (rates) {                  /* if learning rates are given */
    clset->rates[0] = rates[0];
    clset->rates[1] = rates[1];
    clset->rates[2] = rates[2];
  }                             /* copy them to the cluster set */
  if (decays) {                 /* if decay parameters are given */
    clset->decay[0] = decays[0];
    clset->decay[1] = decays[1];
    clset->decay[2] = decays[2];
  }                             /* copy them to the cluster set */
}  /* cls_lrate() */

/*--------------------------------------------------------------------*/

void cls_factors (CLSET *clset, double growth, double shrink)
{                               /* --- set the update factors */
  assert(clset);                /* check the function argument */
  clset->growth =  (growth > 1)                  ? growth : 1;
  clset->shrink = ((shrink > 0) && (shrink < 1)) ? shrink : 1;
  clset->maxfac = clset->growth /(clset->growth +1);
}  /* cls_factors() */

/*--------------------------------------------------------------------*/

void cls_limits (CLSET *clset, double minchg, double maxchg)
{                               /* --- set the update limits */
  assert(clset);                /* check the function argument */
  clset->minchg = (minchg > 0) ? minchg : 0;
  clset->maxchg = (maxchg > 0) ? maxchg : 2;
}  /* cls_limits() */

/*--------------------------------------------------------------------*/

void cls_agg1 (CLSET *clset, DIMID clsid,
               const double *vec, double weight)
{                               /* --- aggregate a data vector */
  if (vec) nst_norm(clset->nst, vec, clset->vec);
  vec = clset->vec;             /* scale given vector to the buffer */
  mat_addvec(clset->cls[clsid].smp, vec, weight);
}  /* cls_agg1() */             /* sum the weighted data vector */

/*--------------------------------------------------------------------*/

DIMID cls_aggr (CLSET *clset, const double *vec, double weight)
{                               /* --- aggregate a data vector */
  DIMID    i;                   /* cluster index, loop variable */
  int      clrn;                /* flag for competitive learning */
  CLUSTER  *c;                  /* to traverse the clusters */
  MATADDFN *add;                /* aggregation function */
  double   x;                   /* adaptation exponent */
  double   msd;                 /* membership degree */

  assert(clset);                /* check the function arguments */
  i = cls_exec(clset,vec,NULL); /* compute degrees of membership */
  vec = clset->vec;             /* get the aggregation function */
  if      (clset->type &  CLS_COVARS)         add = mat_addmpx;
  else if (clset->type & (CLS_VARS|CLS_SIZE)) add = mat_addsvx;
  else                                        add = mat_addvec;

  /* --- alternating estimation and competitive learning --- */
  /* Aggregate the data vectors for the update step. */
  clrn = ((clset->method & CLS_METHOD) == CLS_COMPLRN);
  c    = clset->cls;            /* get the cluster array */
  if (((clset->method & CLS_METHOD) == CLS_ALTOPT) || clrn) {
    x = clset->msexp;           /* get the adaptation exponent */
    if (clset->norm == CLS_HARD) { /* if hard/crisp clustering */
      if (clrn) vec = c[i].dif; /* use diff. vector for comp. learn. */
      add(c[i].smp, vec, weight);} /* sum the weighted data vector */
    else if (x >= 1) {          /* if fuzzy/probabilistic clustering */
      for (i = 0; i < clset->clscnt; i++) {
        msd = c[i].msd;         /* traverse the clusters */
        if (msd <= 0) continue; /* skip cluster with zero membership */
        msd = weight *pow(msd, x); /* compute the data point weight */
        if (clrn) vec = c[i].dif;  /* use diff. for comp. learn. */
        add(c[i].smp, vec, msd);/* sum the data vector weighted */
      } }                       /* with the degree of membership */
    else {                      /* if alt. fuzzifier clustering */
      x = (1-x) /(1+x);         /* compute transformation parameter */
      for (i = 0; i < clset->clscnt; i++) {
        msd = c[i].msd;         /* traverse the clusters */
        msd = (x *(msd-1) +1) *msd *weight;
        if (!(msd > 0)) continue;  /* skip zero membership degrees */
        if (clrn) vec = c[i].dif;  /* use diff. for comp. learn. */
        add(c[i].smp, vec, msd);/* compute the data point weight and */
      }                         /* sum the data vector weighted */
    } }                         /* with the degree of membership */

  /* --- gradient based update --- */
  else {                        /* if (method == CLS_GRADIENT) */
    /* ... to be done ... */
  }

  return i;                     /* return index of best cluster */
}  /* cls_aggr() */

/*--------------------------------------------------------------------*/

void cls_bkprop (CLSET *clset, const double *errs)
{                               /* --- backpropagate errors */
  DIMID   i;                    /* loop variable */
  CLUSTER *c;                   /* to traverse the clusters */
  double  *d, *b, t, v;         /* buffers for computations */

  assert(clset && errs);        /* check the function arguments */
  c = clset->cls;               /* get the cluster array and */
  b = clset->buf;               /* the buffer for computations */
  for (i = 0; i < clset->clscnt; i++) {
    d = c[i].dif;               /* traverse the clusters */
    /* d is the difference vector between input and cluster center. */
    t = errs[i] *clset->drvfn(c[i].d2, clset->rfnps, c[i].msd);
    /* The vector errs contains backpropagated errors per cluster; */
    /* second factor yields d out / d net without inner derivative. */
    if      (clset->type & CLS_COVARS) {
      mat_mulmv(b, c[i].inv, d);/* if adaptable covariances, */
      mat_addmp(c[i].smp, b, t);/* compute derivative terms */
      d = b; }                  /* and get the buffered result */
    else if (clset->type & CLS_VARS) {
      mat_muldv(b, c[i].inv, d);/* if adaptable variances, */
      mat_addsv(c[i].smp, b, t);/* compute derivative terms */
      d = b; }                  /* and get the buffered result */
    else if (clset->type & CLS_SIZE) { /* if adaptable size */
      t *= v = mat_get(c[i].inv, 0, 0);
      mat_inc(c[i].smp, 0, 0, t *c[i].d2 *v);
    }                           /* sum the variance gradients, */
                                /* include the inverse variance */
    vec_add(c[i].sum, clset->inpcnt, c[i].sum, -2 *t, d);
    /* The factor 2 comes from the squared difference, the minus sign */
    /* from the inner derivative (center coordinates are subtracted). */
  }                             /* sum the center gradients */
}  /* cls_bkprop() */

/*--------------------------------------------------------------------*/

double cls_uplain (CLSET *clset)
{                               /* --- plain cluster center update */
  DIMID   i;                    /* loop variable */
  CLUSTER *c;                   /* to traverse the clusters */
  MATRIX  *mat;                 /* exchange buffer for matrices */
  double  max = 0, t;           /* maximal change of a center coord. */

  assert(clset);                /* check the function argument */
  c = clset->cls;               /* get the cluster array */
  for (i = 0; i < clset->clscnt; i++) {
    mat = c[i].smp; c[i].smp = c[i].cov; c[i].cov = mat;
    c[i].ctr = mat_vector(c[i].cov); /* exchange covariance matrix */
    c[i].sum = mat_vector(c[i].smp); /* and aggregation matrix */
    if (mat_getwgt(c[i].cov) >= MINWEIGHT)  /* compute (new) */
         mat_mean (c[i].ctr, c[i].cov);     /* cluster centers*/
    else mat_copyx(c[i].cov, c[i].smp, MAT_VECTOR);
    c[i].move = t = sqrt(vec_sqrdist(c[i].ctr, c[i].sum,clset->inpcnt));
    if (t > max) max = t;       /* compute distance a center moved */
    mat_init(c[i].smp, MAT_VECTOR|MAT_WEIGHT, NULL);
  }                             /* reinit. the cluster aggregation */
  return max;                   /* return the maximal movement */
}  /* cls_uplain() */

/*--------------------------------------------------------------------*/

double cls_update (CLSET *clset, int conly)
{                               /* --- update a set of clusters */
  DIMID   i;                    /* loop variable */
  CLUSTER *c;                   /* to traverse the clusters */
  int     type, m;              /* cluster type, (update) method */
  double  max = 0, t;           /* maximal change of a center coord. */

  assert(clset);                /* check the function argument */
  m    = clset->method & CLS_METHOD;
  type = clset->type;           /* get update method and cluster type */
  if (conly) clset->type = CLS_CENTER;

  /* --- compute new cluster parameters --- */
  clset->rbfgrad = (m == CLS_RBFGRAD);
  switch (m) {                  /* according to the update method */
    case CLS_ALTOPT : altopt  (clset); break;
    case CLS_COMPLRN: complrn (clset); break;
    case CLS_RBFGRAD: rbfgrad (clset); break;
    default         : gradient(clset); break;
  }                             /* call the approp. update function */
  /* After these calls c->cov contains the new (unscaled) covariance */
  /* matrix, while c->smp is the old covariance matrix. Analogously, */
  /* c->ctr is the new cluster center, while c->sum is the old.      */
  m = ((clset->fwexp > 0) && (clset->fwexp != 1))
    || (clset->regps[3] < -1) || (clset->regps[3] > 0);

  if (clset->method & CLS_MODIFIER)
    neural(clset, m);           /* neural network inspired update */
  /* The neural network inspired update is carried out before any */
  /* regularization, because its execution may destroy effects of */
  /* a regularization, but not all methods can be executed again. */
  /* (That is, some regularization methods are not idempotent.)   */

  /* --- regularization and scaling --- */
  cls_vars(clset, m);           /* compute the isotropic variances */
  if ((clset->type & (CLS_VARS|CLS_COVARS))
  && ((clset->fwexp    <= 0) || (clset->fwexp    >= 1))
  && ((clset->regps[3] < -1) || (clset->regps[3] >  0)))
    regshape(clset);            /* regularize shape (axes lengths) */
  if ((clset->type & CLS_SIZE) && !(clset->type & CLS_JOINT)
  && ((clset->fwexp <= 0) || (clset->fwexp == 1)))
    regsize(clset);             /* regularize the cluster sizes */
  /* For a shape and size regularization only the isotropic variances */
  /* needs to be known, hence gauging and resizing can be done later. */
  cls_gauge(clset);             /* measure the cluster sizes */
  cls_resize(clset, 1);         /* and resize the clusters */
  if ((clset->type & CLS_WEIGHT) && !(clset->type & CLS_JOINT)
  && ((clset->regps[4] < -1) || (clset->regps[4] > 0)))
    regwgts(clset);             /* regularize the cluster weights */
  if (clset->method & CLS_UNIT)
    regctrs(clset);             /* regularize the cluster centers */

  /* --- determine maximal change --- */
  c = clset->cls;               /* get the cluster array */
  for (i = 0; i < clset->clscnt; i++) {
    t = vec_diff(c[i].ctr, c[i].sum, clset->inpcnt);
    if (t > max) max = t;       /* compare the cluster centers and */
  }                             /* determine the maximal change */

  /* --- finalize update --- */
  if (conly) clset->type = type;/* restore the cluster type */
  cls_ftwgts(clset);            /* compute the feature weights */
  cls_invert(clset);            /* and the inverse matrices */
  if ((clset->type &  CLS_JOINT)
  &&  (clset->type & (CLS_COVARS|CLS_VARS|CLS_SIZE)))
    cls_distrib(clset);         /* disribute joint (co)variances */
  cls_reinit(clset);            /* reinit. aggregation matrices */
  clset->setup = 1;             /* cluster set has been set up */
  return max;                   /* return the maximal change */
}  /* cls_update() */

/*--------------------------------------------------------------------*/

void cls_sort (CLSET *clset)
{                               /* --- sort clusters by their centers */
  DIMID   i, s, d, n;           /* loop variables, indices */
  DIMID   h;                    /* increment for index */
  CLUSTER x;                    /* exchange buffer */

  assert(clset);                /* check the function argument */
  n = clset->clscnt;            /* get the number of clusters */
  for (h = n; h > 1; ) {        /* traverse the increments */
    h = (h < 5) ? 1 : ((h *5 -1) /11);   /* (Shell sort) */
    for (i = h-1; ++i < n; ) {  /* traverse the clusters */
      x = clset->cls[d = i];    /* note the cluster to insert and do */
      do {                      /* insertion sort with increment h */
        s = d-h;                /* to find the insertion position */
        if (vec_cmp(clset->cls[s].ctr, x.ctr, clset->inpcnt) <= 0)
          break;                /* shift up all clusters that must */
        clset->cls[d] = clset->cls[s];   /* follow the current one */
      } while ((d = s) >= h);   /* while another shift is possible */
      clset->cls[d] = x;        /* store the cluster to insert */
    }                           /* at the insertion position */
  }
}  /* cls_sort() */
